@@ -1,17 +1,30 @@
+from __future__ import annotations
+
 from abc import ABCMeta, abstractmethod
 import asyncio
 from decimal import Decimal
-import hashlib
+from io import BytesIO, SEEK_END
 import logging
 from pathlib import Path
 import pickle
+import re
 import signal
 import time
 from typing import (
-    Any, Collection, Optional, Type, Union,
-    Mapping, MutableMapping, Dict,
-    Set, FrozenSet,
-    Sequence, MutableSequence, Tuple,
+    Any,
+    AsyncIterator,
+    Collection,
+    Dict,
+    FrozenSet,
+    Optional,
+    Mapping,
+    MutableMapping,
+    MutableSequence,
+    Sequence,
+    Set,
+    Tuple,
+    Union,
+    TYPE_CHECKING,
     cast,
 )
 
@@ -24,28 +37,26 @@ import zmq, zmq.asyncio
 
 from ai.backend.common import msgpack, redis
 from ai.backend.common.docker import (
+    ImageRef,
     MIN_KERNELSPEC,
     MAX_KERNELSPEC,
 )
 from ai.backend.common.logging import BraceStyleAdapter
-from ai.backend.common.monitor import DummyStatsMonitor, DummyErrorMonitor
-from ai.backend.common.plugin import install_plugins
 from ai.backend.common.types import (
     aobject,
     # TODO: eliminate use of ContainerId
     ContainerId, KernelId,
     DeviceName, SlotName,
+    AutoPullBehavior, ImageRegistry,
     KernelCreationConfig,
     KernelCreationResult,
-    MetricKey, MetricValue,
     Sentinel,
 )
+from ai.backend.common.utils import current_loop
+from ai.backend.common.plugin.monitor import ErrorPluginContext, StatsPluginContext
 from . import __version__ as VERSION
 from .defs import ipc_base_path
 from .kernel import AbstractKernel
-from .utils import (
-    current_loop,
-)
 from .resources import (
     AbstractComputeDevice,
     AbstractComputePlugin,
@@ -60,10 +71,14 @@ from .types import (
     ContainerLifecycleEvent,
     LifecycleEvent,
 )
+from .utils import generate_agent_id
+
+if TYPE_CHECKING:
+    from ai.backend.common.etcd import AsyncEtcd
 
 log = BraceStyleAdapter(logging.getLogger('ai.backend.agent.agent'))
 
-_sentinel = Sentinel()
+_sentinel = Sentinel.TOKEN
 
 ACTIVE_STATUS_SET = frozenset([
     ContainerStatus.RUNNING,
@@ -87,7 +102,7 @@ class RestartTracker:
 
 @attr.s(auto_attribs=True, slots=True)
 class ComputerContext:
-    klass: Type[AbstractComputePlugin]
+    instance: AbstractComputePlugin
     devices: Collection[AbstractComputeDevice]
     alloc_map: AbstractAllocMap
 
@@ -95,7 +110,8 @@ class ComputerContext:
 class AbstractAgent(aobject, metaclass=ABCMeta):
 
     loop: asyncio.AbstractEventLoop
-    config: Mapping[str, Any]
+    local_config: Mapping[str, Any]
+    etcd: AsyncEtcd
     agent_id: str
     kernel_registry: MutableMapping[KernelId, AbstractKernel]
     computers: MutableMapping[str, ComputerContext]
@@ -114,28 +130,39 @@ class AbstractAgent(aobject, metaclass=ABCMeta):
     stat_sync_sockpath: Path
     stat_sync_task: asyncio.Task
 
-    def __init__(self, config: Mapping[str, Any]) -> None:
+    stats_monitor: StatsPluginContext
+    error_monitor: ErrorPluginContext
+
+    def __init__(
+        self,
+        etcd: AsyncEtcd,
+        local_config: Mapping[str, Any],
+        *,
+        stats_monitor: StatsPluginContext,
+        error_monitor: ErrorPluginContext,
+        skip_initial_scan: bool = False,
+    ) -> None:
+        self._skip_initial_scan = skip_initial_scan
         self.loop = current_loop()
-        self.config = config
-        self.agent_id = hashlib.md5(__file__.encode('utf-8')).hexdigest()[:12]
+        self.etcd = etcd
+        self.local_config = local_config
+        self.agent_id = generate_agent_id(__file__)
         self.kernel_registry = {}
         self.computers = {}
         self.images = {}  # repoTag -> digest
         self.terminating_kernels = set()
         self.restarting_kernels = {}
-        self.stat_ctx = StatContext(self, mode=StatModes(config['container']['stats-type']))
+        self.stat_ctx = StatContext(
+            self, mode=StatModes(local_config['container']['stats-type']),
+        )
         self.timer_tasks = []
         self.port_pool = set(range(
-            config['container']['port-range'][0],
-            config['container']['port-range'][1] + 1,
+            local_config['container']['port-range'][0],
+            local_config['container']['port-range'][1] + 1,
         ))
-        self.stats_monitor = DummyStatsMonitor()
-        self.error_monitor = DummyErrorMonitor()
-        plugins = [
-            'stats_monitor',
-            'error_monitor'
-        ]
-        install_plugins(plugins, self, 'attr', self.config)
+        self.stats_monitor = stats_monitor
+        self.error_monitor = error_monitor
+        self._rx_distro = re.compile(r"\.([a-z-]+\d+\.\d+)\.")
 
     async def __ainit__(self) -> None:
         """
@@ -146,34 +173,33 @@ class AbstractAgent(aobject, metaclass=ABCMeta):
         self.container_lifecycle_queue = asyncio.Queue()
         self.producer_lock = asyncio.Lock()
         self.redis_producer_pool = await redis.connect_with_retries(
-            self.config['redis']['addr'].as_sockaddr(),
+            self.local_config['redis']['addr'].as_sockaddr(),
             db=4,  # REDIS_STREAM_DB in gateway.defs
-            password=(self.config['redis']['password']
-                      if self.config['redis']['password'] else None),
+            password=(self.local_config['redis']['password']
+                      if self.local_config['redis']['password'] else None),
             encoding=None,
         )
         self.redis_stat_pool = await redis.connect_with_retries(
-            self.config['redis']['addr'].as_sockaddr(),
+            self.local_config['redis']['addr'].as_sockaddr(),
             db=0,  # REDIS_STAT_DB in backend.ai-manager
-            password=(self.config['redis']['password']
-                      if self.config['redis']['password'] else None),
+            password=(self.local_config['redis']['password']
+                      if self.local_config['redis']['password'] else None),
             encoding='utf8',
         )
 
         ipc_base_path.mkdir(parents=True, exist_ok=True)
         self.zmq_ctx = zmq.asyncio.Context()
 
-        computers, self.slots = await self.detect_resources(
-            self.config['resource'],
-            self.config['plugins'])
-        for name, klass in computers.items():
-            devices = await klass.list_devices()
-            alloc_map = await klass.create_alloc_map()
-            self.computers[name] = ComputerContext(klass, devices, alloc_map)
+        computers, self.slots = await self.detect_resources()
+        for name, computer in computers.items():
+            devices = await computer.list_devices()
+            alloc_map = await computer.create_alloc_map()
+            self.computers[name] = ComputerContext(computer, devices, alloc_map)
 
-        self.images = await self.scan_images()
-        self.timer_tasks.append(aiotools.create_timer(self._scan_images_wrapper, 20.0))
-        await self.scan_running_kernels()
+        if not self._skip_initial_scan:
+            self.images = await self.scan_images()
+            self.timer_tasks.append(aiotools.create_timer(self._scan_images_wrapper, 20.0))
+            await self.scan_running_kernels()
 
         # Prepare stat collector tasks.
         self.timer_tasks.append(aiotools.create_timer(self.collect_node_stat, 5.0))
@@ -229,11 +255,17 @@ class AbstractAgent(aobject, metaclass=ABCMeta):
         """
         Send an event to the manager(s).
         """
-        _log = log.debug if event_name == 'instance_heartbeat' else log.info
-        _log('produce_event({0})', event_name)
+        if self.local_config['debug']['log-heartbeats']:
+            _log = log.debug if event_name == 'instance_heartbeat' else log.info
+        else:
+            _log = (lambda *args: None) if event_name == 'instance_heartbeat' else log.info
+        if event_name.startswith('kernel_') and len(args) > 0:
+            _log('produce_event({0}, k:{1})', event_name, args[0])
+        else:
+            _log('produce_event({0})', event_name)
         encoded_event = msgpack.packb({
             'event_name': event_name,
-            'agent_id': self.config['agent']['id'],
+            'agent_id': self.local_config['agent']['id'],
             'args': args,
         })
         async with self.producer_lock:
@@ -250,22 +282,22 @@ class AbstractAgent(aobject, metaclass=ABCMeta):
         """
         res_slots = {}
         for cctx in self.computers.values():
-            for slot_key, slot_type in cctx.klass.slot_types:
+            for slot_key, slot_type in cctx.instance.slot_types:
                 res_slots[slot_key] = (
                     slot_type,
                     str(self.slots.get(slot_key, 0)),
                 )
         agent_info = {
-            'ip': str(self.config['agent']['rpc-listen-addr'].host),
-            'region': self.config['agent']['region'],
-            'scaling_group': self.config['agent']['scaling-group'],
-            'addr': f"tcp://{self.config['agent']['rpc-listen-addr']}",
+            'ip': str(self.local_config['agent']['rpc-listen-addr'].host),
+            'region': self.local_config['agent']['region'],
+            'scaling_group': self.local_config['agent']['scaling-group'],
+            'addr': f"tcp://{self.local_config['agent']['rpc-listen-addr']}",
             'resource_slots': res_slots,
             'version': VERSION,
             'compute_plugins': {
                 key: {
-                    'version': computer.klass.get_version(),
-                    **(await computer.klass.extra_info())
+                    'version': computer.instance.get_version(),
+                    **(await computer.instance.extra_info())
                 }
                 for key, computer in self.computers.items()
             },
@@ -279,20 +311,72 @@ class AbstractAgent(aobject, metaclass=ABCMeta):
             log.warning('event dispatch timeout: instance_heartbeat')
         except Exception:
             log.exception('instance_heartbeat failure')
-            self.error_monitor.capture_exception()
+            await self.error_monitor.capture_exception()
+
+    async def collect_logs(
+        self,
+        kernel_id: KernelId,
+        container_id: str,
+        async_log_iterator: AsyncIterator[bytes],
+    ) -> None:
+        chunk_size = self.local_config['agent']['container-logs']['chunk-size']
+        log_key = f'containerlog.{container_id}'
+        log_length = 0
+        chunk_buffer = BytesIO()
+        chunk_length = 0
+        try:
+            async for fragment in async_log_iterator:
+                fragment_length = len(fragment)
+                chunk_buffer.write(fragment)
+                chunk_length += fragment_length
+                log_length += fragment_length
+                while chunk_length >= chunk_size:
+                    cb = chunk_buffer.getbuffer()
+                    stored_chunk = bytes(cb[:chunk_size])
+                    await redis.execute_with_retries(
+                        lambda: self.redis_producer_pool.rpush(
+                            log_key, stored_chunk)
+                    )
+                    remaining = cb[chunk_size:]
+                    chunk_length = len(remaining)
+                    next_chunk_buffer = BytesIO(remaining)
+                    next_chunk_buffer.seek(0, SEEK_END)
+                    del remaining, cb
+                    chunk_buffer.close()
+                    chunk_buffer = next_chunk_buffer
+            assert chunk_length < chunk_size
+            if chunk_length > 0:
+                await redis.execute_with_retries(
+                    lambda: self.redis_producer_pool.rpush(
+                        log_key, chunk_buffer.getvalue())
+                )
+        finally:
+            chunk_buffer.close()
+        # Keep the log for at most one hour in Redis.
+        # This is just a safety measure to prevent memory leak in Redis
+        # for cases when the event delivery has failed or processing
+        # the log data has failed.
+        await redis.execute_with_retries(
+            lambda: self.redis_producer_pool.expire(log_key, 3600.0)
+        )
+        await self.produce_event(
+            'kernel_log', str(kernel_id), container_id
+        )
 
     async def collect_node_stat(self, interval: float):
-        log.debug('collecting node statistics')
+        if self.local_config['debug']['log-stats']:
+            log.debug('collecting node statistics')
         try:
             await self.stat_ctx.collect_node_stat()
         except asyncio.CancelledError:
             pass
         except Exception:
             log.exception('unhandled exception while syncing node stats')
-            self.error_monitor.capture_exception()
+            await self.error_monitor.capture_exception()
 
     async def collect_container_stat(self, interval: float):
-        log.debug('collecting container statistics')
+        if self.local_config['debug']['log-stats']:
+            log.debug('collecting container statistics')
         try:
             updated_kernel_ids = []
             for kernel_id, kernel_obj in [*self.kernel_registry.items()]:
@@ -309,7 +393,7 @@ class AbstractAgent(aobject, metaclass=ABCMeta):
             pass
         except Exception:
             log.exception('unhandled exception while syncing container stats')
-            self.error_monitor.capture_exception()
+            await self.error_monitor.capture_exception()
 
     async def _handle_start_event(self, ev: ContainerLifecycleEvent) -> None:
         kernel_obj = self.kernel_registry.get(ev.kernel_id)
@@ -347,7 +431,7 @@ class AbstractAgent(aobject, metaclass=ABCMeta):
             result = await self.destroy_kernel(ev.kernel_id, ev.container_id)
         except Exception:
             log.exception('unhandled exception while processing DESTROY event')
-            self.error_monitor.capture_exception()
+            await self.error_monitor.capture_exception()
         finally:
             if ev.done_event is not None:
                 ev.done_event.set()
@@ -366,13 +450,13 @@ class AbstractAgent(aobject, metaclass=ABCMeta):
             )
         except Exception:
             log.exception('unhandled exception while processing CLEAN event')
-            self.error_monitor.capture_exception()
+            await self.error_monitor.capture_exception()
         finally:
             try:
                 kernel_obj = self.kernel_registry.get(ev.kernel_id)
                 if kernel_obj is not None:
                     # Restore used ports to the port pool.
-                    port_range = self.config['container']['port-range']
+                    port_range = self.local_config['container']['port-range']
                     restored_ports = [*filter(
                         lambda p: port_range[0] <= p <= port_range[1],
                         kernel_obj['host_ports']
@@ -398,21 +482,22 @@ class AbstractAgent(aobject, metaclass=ABCMeta):
                     )
 
     async def process_lifecycle_events(self) -> None:
-        loop = current_loop()
         while True:
             ev = await self.container_lifecycle_queue.get()
             if isinstance(ev, Sentinel):
                 with open(ipc_base_path / f'last_registry.{self.agent_id}.dat', 'wb') as f:
                     pickle.dump(self.kernel_registry, f)
                 return
-            log.info('lifecycle event: {!r}', ev)
+            # attr currently does not support customizing getstate/setstate dunder methods
+            # until the next release.
+            log.info(f'lifecycle event: {ev!r}')
             try:
                 if ev.event == LifecycleEvent.START:
-                    loop.create_task(self._handle_start_event(ev))
+                    asyncio.create_task(self._handle_start_event(ev))
                 elif ev.event == LifecycleEvent.DESTROY:
-                    loop.create_task(self._handle_destroy_event(ev))
+                    asyncio.create_task(self._handle_destroy_event(ev))
                 elif ev.event == LifecycleEvent.CLEAN:
-                    loop.create_task(self._handle_clean_event(ev))
+                    asyncio.create_task(self._handle_clean_event(ev))
                 else:
                     log.warning('unsupported lifecycle event: {!r}', ev)
             except Exception:
@@ -472,7 +557,7 @@ class AbstractAgent(aobject, metaclass=ABCMeta):
                 computer_set.alloc_map.clear()
             for kernel_id, container in (await self.enumerate_containers()):
                 for computer_set in self.computers.values():
-                    await computer_set.klass.restore_from_container(
+                    await computer_set.instance.restore_from_container(
                         container,
                         computer_set.alloc_map,
                     )
@@ -551,14 +636,11 @@ class AbstractAgent(aobject, metaclass=ABCMeta):
                        for kernel_id in kernel_ids]
             await asyncio.gather(*waiters)
 
-    @staticmethod
     @abstractmethod
-    async def detect_resources(resource_configs: Mapping[str, Any],
-                               plugin_configs: Mapping[str, Any]) \
-                               -> Tuple[
-                                   Mapping[DeviceName, Type[AbstractComputePlugin]],
-                                   Mapping[SlotName, Decimal]
-                               ]:
+    async def detect_resources(self) -> Tuple[
+        Mapping[DeviceName, AbstractComputePlugin],
+        Mapping[SlotName, Decimal]
+    ]:
         """
         Scan and define the amount of available resource slots in this node.
         """
@@ -574,6 +656,20 @@ class AbstractAgent(aobject, metaclass=ABCMeta):
     async def _scan_images_wrapper(self, interval: float) -> None:
         self.images = await self.scan_images()
 
+    @abstractmethod
+    async def pull_image(self, image_ref: ImageRef, registry_conf: ImageRegistry) -> None:
+        '''
+        Pull the given image from the given registry.
+        '''
+
+    @abstractmethod
+    async def check_image(self, image_ref: ImageRef, image_id: str, auto_pull: AutoPullBehavior) -> bool:
+        '''
+        Check the availability of the image and return a boolean flag that indicates whether
+        the agent should try pulling the image from a registry.
+        '''
+        return False
+
     async def scan_running_kernels(self) -> None:
         """
         Scan currently running kernels and recreate the kernel objects in
@@ -583,7 +679,7 @@ class AbstractAgent(aobject, metaclass=ABCMeta):
             with open(ipc_base_path / f'last_registry.{self.agent_id}.dat', 'rb') as f:
                 self.kernel_registry = pickle.load(f)
                 for kernel_obj in self.kernel_registry.values():
-                    kernel_obj.agent_config = self.config
+                    kernel_obj.agent_config = self.local_config
                     if kernel_obj.runner is not None:
                         await kernel_obj.runner.__ainit__()
         except FileNotFoundError:
@@ -602,7 +698,7 @@ class AbstractAgent(aobject, metaclass=ABCMeta):
                             self.port_pool.discard(p.host_port)
                     # Restore compute resources.
                     for computer_set in self.computers.values():
-                        await computer_set.klass.restore_from_container(
+                        await computer_set.instance.restore_from_container(
                             container,
                             computer_set.alloc_map,
                         )
@@ -629,16 +725,22 @@ class AbstractAgent(aobject, metaclass=ABCMeta):
                      dict(computer_ctx.alloc_map.allocations))
 
     @abstractmethod
-    async def create_kernel(self, kernel_id: KernelId, config: KernelCreationConfig, *,
-                            restarting: bool = False) -> KernelCreationResult:
+    async def create_kernel(
+        self,
+        kernel_id: KernelId,
+        config: KernelCreationConfig,
+        *,
+        restarting: bool = False,
+    ) -> KernelCreationResult:
         """
         Create a new kernel.
 
         Things to do:
         * ``await self.produce_event('kernel_preparing', kernel_id)``
-        * Check availability of kernel image.
-        * ``await self.produce_event('kernel_pulling', kernel_id)``
-        * Pull the kernel image if image is not ready in the agent. (optional; may return error)
+        * Check availability of kernel image using ``await self.check_image(...)``.
+          - If it returns True:
+          - ``await self.produce_event('kernel_pulling', kernel_id)``
+          - Pull the kernel image using ``await self.pull_image(...)``.
         * ``await self.produce_event('kernel_creating', kernel_id)``
         * Create KernelResourceSpec object or read it from kernel's internal configuration cache.
           Use *restarting* argument to detect if the kernel is being created or restarted.
@@ -659,7 +761,7 @@ class AbstractAgent(aobject, metaclass=ABCMeta):
         self,
         kernel_id: KernelId,
         container_id: Optional[ContainerId],
-    ) -> Optional[Mapping[MetricKey, MetricValue]]:
+    ) -> None:
         """
         Initiate destruction of the kernel.
 
@@ -701,7 +803,11 @@ class AbstractAgent(aobject, metaclass=ABCMeta):
                 destroy_event=asyncio.Event(),
                 done_event=asyncio.Event())
             self.restarting_kernels[kernel_id] = tracker
-        config_dir = (self.config['container']['scratch-root'] / str(kernel_id) / 'config').resolve()
+        config_dir = (
+            self.local_config['container']['scratch-root'] /
+            str(kernel_id) /
+            'config'
+        ).resolve()
         with open(config_dir / 'kconfig.dat', 'rb') as fb:
             existing_config = pickle.load(fb)
             new_config = cast(KernelCreationConfig, {**existing_config, **new_config})
