@@ -7,8 +7,12 @@ import platform
 from typing import (
     cast,
     Any,
-    Collection, Mapping,
-    Sequence, List,
+    Collection,
+    Dict,
+    List,
+    Mapping,
+    Optional,
+    Sequence,
 )
 
 import aiohttp
@@ -17,7 +21,7 @@ from aiodocker.exceptions import DockerError
 import psutil
 
 from ai.backend.common.logging import BraceStyleAdapter
-from ai.backend.common.utils import nmget
+from ai.backend.common.utils import current_loop, nmget
 from ai.backend.common.types import (
     DeviceName, DeviceId,
     DeviceModelInfo,
@@ -39,12 +43,49 @@ from ..stats import (
     StatContext, NodeMeasurement, ContainerMeasurement,
     StatModes, MetricTypes, Measurement,
 )
-from ..utils import read_sysfs, current_loop
+from ..utils import read_sysfs
 from ..vendor.linux import libnuma
 
 log = BraceStyleAdapter(logging.getLogger(__name__))
 
 
+async def fetch_api_stats(container: DockerContainer) -> Optional[Dict[str, Any]]:
+    try:
+        short_cid = container._id[:7]
+        ret = await container.stats(stream=False)  # TODO: cache
+    except RuntimeError as e:
+        msg = str(e.args[0]).lower()
+        if 'event loop is closed' in msg or 'session is closed' in msg:
+            return None
+        raise
+    except (DockerError, aiohttp.ClientError) as e:
+        log.error(
+            'cannot read stats (cid:{}): client error: {!r}.',
+            short_cid, e,
+        )
+        return None
+    else:
+        # aiodocker 0.16 or later returns a list of dict, even when not streaming.
+        if isinstance(ret, list):
+            if not ret:
+                # The API may return an empty result upon container termination.
+                return None
+            ret = ret[0]
+        # The API may return an invalid or empty result upon container termination.
+        if ret is None or not isinstance(ret, dict):
+            log.warning(
+                'cannot read stats (cid:{}): got an empty result: {}',
+                short_cid, ret,
+            )
+            return None
+        if (
+            ret['read'].startswith('0001-01-01') or
+            ret['preread'].startswith('0001-01-01')
+        ):
+            return None
+        return ret
+
+
 # Pseudo-plugins for intrinsic devices (CPU and the main memory)
 
 class CPUDevice(AbstractComputeDevice):
@@ -52,18 +93,28 @@ class CPUDevice(AbstractComputeDevice):
 
 
 class CPUPlugin(AbstractComputePlugin):
-    '''
+    """
     Represents the CPU.
-    '''
+    """
+
+    config_watch_enabled = False
 
     key = DeviceName('cpu')
     slot_types = [
         (SlotName('cpu'), SlotTypes.COUNT)
     ]
 
-    @classmethod
-    async def list_devices(cls) -> Collection[CPUDevice]:
-        num_cores = libnuma.get_available_cores()
+    async def init(self, context: Any = None) -> None:
+        pass
+
+    async def cleanup(self) -> None:
+        pass
+
+    async def update_plugin_config(self, new_plugin_config: Mapping[str, Any]) -> None:
+        pass
+
+    async def list_devices(self) -> Collection[CPUDevice]:
+        cores = await libnuma.get_available_cores()
         overcommit_factor = int(os.environ.get('BACKEND_CPU_OVERCOMMIT_FACTOR', '1'))
         assert 1 <= overcommit_factor <= 4
         return [
@@ -74,30 +125,26 @@ class CPUPlugin(AbstractComputePlugin):
                 memory_size=0,
                 processing_units=1 * overcommit_factor,
             )
-            for core_idx in sorted(num_cores)
+            for core_idx in sorted(cores)
         ]
 
-    @classmethod
-    async def available_slots(cls) -> Mapping[SlotName, Decimal]:
-        devices = await cls.list_devices()
+    async def available_slots(self) -> Mapping[SlotName, Decimal]:
+        devices = await self.list_devices()
         return {
             SlotName('cpu'): Decimal(sum(dev.processing_units for dev in devices)),
         }
 
-    @classmethod
-    def get_version(cls) -> str:
+    def get_version(self) -> str:
         return __version__
 
-    @classmethod
-    async def extra_info(cls) -> Mapping[str, str]:
+    async def extra_info(self) -> Mapping[str, str]:
         return {
             'agent_version': __version__,
             'machine': platform.machine(),
             'os_type': platform.system(),
         }
 
-    @classmethod
-    async def gather_node_measures(cls, ctx: StatContext) -> Sequence[NodeMeasurement]:
+    async def gather_node_measures(self, ctx: StatContext) -> Sequence[NodeMeasurement]:
         _cstat = psutil.cpu_times(True)
         q = Decimal('0.000')
         total_cpu_used = cast(Decimal,
@@ -123,9 +170,11 @@ class CPUPlugin(AbstractComputePlugin):
             ),
         ]
 
-    @classmethod
-    async def gather_container_measures(cls, ctx: StatContext, container_ids: Sequence[str]) \
-            -> Sequence[ContainerMeasurement]:
+    async def gather_container_measures(
+        self,
+        ctx: StatContext,
+        container_ids: Sequence[str],
+    ) -> Sequence[ContainerMeasurement]:
 
         async def sysfs_impl(container_id):
             cpu_prefix = f'/sys/fs/cgroup/cpuacct/docker/{container_id}/'
@@ -138,24 +187,10 @@ class CPUPlugin(AbstractComputePlugin):
             return cpu_used
 
         async def api_impl(container_id):
-            try:
-                container = DockerContainer(ctx.agent.docker, id=container_id)
-                ret = await container.stats(stream=False)  # TODO: cache
-            except RuntimeError as e:
-                msg = str(e.args[0]).lower()
-                if 'event loop is closed' in msg or 'session is closed' in msg:
-                    return None
-                raise
-            except (DockerError, aiohttp.ClientError):
-                short_cid = container._id[:7]
-                log.warning(f'cannot read stats: Docker stats API error for {short_cid}.')
+            container = DockerContainer(ctx.agent.docker, id=container_id)
+            ret = await fetch_api_stats(container)
+            if ret is None:
                 return None
-            else:
-                # API returned successfully but actually the result may be empty!
-                if ret is None or not isinstance(ret, dict):
-                    return None
-                if ret['preread'].startswith('0001-01-01'):
-                    return None
             cpu_used = nmget(ret, 'cpu_stats.cpu_usage.total_usage', 0) / 1e6
             return cpu_used
 
@@ -191,23 +226,21 @@ class CPUPlugin(AbstractComputePlugin):
             ),
         ]
 
-    @classmethod
-    async def create_alloc_map(cls) -> AbstractAllocMap:
-        devices = await cls.list_devices()
+    async def create_alloc_map(self) -> AbstractAllocMap:
+        devices = await self.list_devices()
         return DiscretePropertyAllocMap(
             devices=devices,
             prop_func=lambda dev: dev.processing_units)
 
-    @classmethod
-    async def get_hooks(cls, distro: str, arch: str) -> Sequence[Path]:
+    async def get_hooks(self, distro: str, arch: str) -> Sequence[Path]:
         # TODO: move the sysconf hook in libbaihook.so here
         return []
 
-    @classmethod
-    async def generate_docker_args(cls,
-                                   docker: Docker,
-                                   device_alloc) \
-                                  -> Mapping[str, Any]:
+    async def generate_docker_args(
+        self,
+        docker: Docker,
+        device_alloc,
+    ) -> Mapping[str, Any]:
         cores = [*map(int, device_alloc['cpu'].keys())]
         sorted_core_ids = [*map(str, sorted(cores))]
         return {
@@ -220,9 +253,11 @@ class CPUPlugin(AbstractComputePlugin):
             }
         }
 
-    @classmethod
-    async def restore_from_container(cls, container: Container,
-                                     alloc_map: AbstractAllocMap) -> None:
+    async def restore_from_container(
+        self,
+        container: Container,
+        alloc_map: AbstractAllocMap,
+    ) -> None:
         assert isinstance(alloc_map, DiscretePropertyAllocMap)
         # Docker does not return the original cpuset.... :(
         # We need to read our own records.
@@ -234,17 +269,20 @@ class CPUPlugin(AbstractComputePlugin):
                 resource_spec.allocations[DeviceName('cpu')][SlotName('cpu')],
         })
 
-    @classmethod
-    async def get_attached_devices(cls, device_alloc: Mapping[SlotName, Mapping[DeviceId, Decimal]]) \
-                                   -> Sequence[DeviceModelInfo]:
+    async def get_attached_devices(
+        self,
+        device_alloc: Mapping[SlotName,
+        Mapping[DeviceId, Decimal]],
+    ) -> Sequence[DeviceModelInfo]:
         device_ids = [*device_alloc[SlotName('cpu')].keys()]
-        available_devices = await cls.list_devices()
+        available_devices = await self.list_devices()
         attached_devices: List[DeviceModelInfo] = []
         for device in available_devices:
             if device.device_id in device_ids:
                 attached_devices.append({
                     'device_id': device.device_id,
                     'model_name': '',
+                    'data': {'cores': len(device_ids)},
                 })
         return attached_devices
 
@@ -254,20 +292,30 @@ class MemoryDevice(AbstractComputeDevice):
 
 
 class MemoryPlugin(AbstractComputePlugin):
-    '''
+    """
     Represents the main memory.
 
     When collecting statistics, it also measures network and I/O usage
     in addition to the memory usage.
-    '''
+    """
+
+    config_watch_enabled = False
 
     key = DeviceName('mem')
     slot_types = [
         (SlotName('mem'), SlotTypes.BYTES)
     ]
 
-    @classmethod
-    async def list_devices(cls) -> Collection[MemoryDevice]:
+    async def init(self, context: Any = None) -> None:
+        pass
+
+    async def cleanup(self) -> None:
+        pass
+
+    async def update_plugin_config(self, new_plugin_config: Mapping[str, Any]) -> None:
+        pass
+
+    async def list_devices(self) -> Collection[MemoryDevice]:
         # TODO: support NUMA?
         memory_size = psutil.virtual_memory().total
         return [MemoryDevice(
@@ -278,23 +326,19 @@ class MemoryPlugin(AbstractComputePlugin):
             processing_units=0,
         )]
 
-    @classmethod
-    async def available_slots(cls) -> Mapping[SlotName, Decimal]:
-        devices = await cls.list_devices()
+    async def available_slots(self) -> Mapping[SlotName, Decimal]:
+        devices = await self.list_devices()
         return {
             SlotName('mem'): Decimal(sum(dev.memory_size for dev in devices)),
         }
 
-    @classmethod
-    def get_version(cls) -> str:
+    def get_version(self) -> str:
         return __version__
 
-    @classmethod
-    async def extra_info(cls) -> Mapping[str, str]:
+    async def extra_info(self) -> Mapping[str, str]:
         return {}
 
-    @classmethod
-    async def gather_node_measures(cls, ctx: StatContext) -> Sequence[NodeMeasurement]:
+    async def gather_node_measures(self, ctx: StatContext) -> Sequence[NodeMeasurement]:
         _mstat = psutil.virtual_memory()
         total_mem_used_bytes = Decimal(_mstat.total - _mstat.available)
         total_mem_capacity_bytes = Decimal(_mstat.total)
@@ -356,8 +400,7 @@ class MemoryPlugin(AbstractComputePlugin):
             ),
         ]
 
-    @classmethod
-    async def gather_container_measures(cls, ctx: StatContext, container_ids: Sequence[str]) \
+    async def gather_container_measures(self, ctx: StatContext, container_ids: Sequence[str]) \
             -> Sequence[ContainerMeasurement]:
 
         def get_scratch_size(container_id: str) -> int:
@@ -366,7 +409,7 @@ class MemoryPlugin(AbstractComputePlugin):
                     break
             else:
                 return 0
-            work_dir = ctx.agent.config['container']['scratch-root'] / str(kernel_id) / 'work'
+            work_dir = ctx.agent.local_config['container']['scratch-root'] / str(kernel_id) / 'work'
             total_size = 0
             for path in work_dir.rglob('*'):
                 if path.is_symlink():
@@ -408,24 +451,10 @@ class MemoryPlugin(AbstractComputePlugin):
             return mem_cur_bytes, io_read_bytes, io_write_bytes, scratch_sz
 
         async def api_impl(container_id):
-            try:
-                container = DockerContainer(ctx.agent.docker, id=container_id)
-                ret = await container.stats(stream=False)  # TODO: cache
-            except RuntimeError as e:
-                msg = str(e.args[0]).lower()
-                if 'event loop is closed' in msg or 'session is closed' in msg:
-                    return None
-                raise
-            except (DockerError, aiohttp.ClientError):
-                short_cid = container._id[:7]
-                log.warning(f'cannot read stats: Docker stats API error for {short_cid}.')
+            container = DockerContainer(ctx.agent.docker, id=container_id)
+            ret = await fetch_api_stats(container)
+            if ret is None:
                 return None
-            else:
-                # API returned successfully but actually the result may be empty!
-                if ret is None or not isinstance(ret, dict):
-                    return None
-                if ret['preread'].startswith('0001-01-01'):
-                    return None
             mem_cur_bytes = nmget(ret, 'memory_stats.usage', 0)
             io_read_bytes = 0
             io_write_bytes = 0
@@ -494,22 +523,20 @@ class MemoryPlugin(AbstractComputePlugin):
             ),
         ]
 
-    @classmethod
-    async def create_alloc_map(cls) -> AbstractAllocMap:
-        devices = await cls.list_devices()
+    async def create_alloc_map(self) -> AbstractAllocMap:
+        devices = await self.list_devices()
         return DiscretePropertyAllocMap(
             devices=devices,
             prop_func=lambda dev: dev.memory_size)
 
-    @classmethod
-    async def get_hooks(cls, distro: str, arch: str) -> Sequence[Path]:
+    async def get_hooks(self, distro: str, arch: str) -> Sequence[Path]:
         return []
 
-    @classmethod
-    async def generate_docker_args(cls,
-                                   docker: Docker,
-                                   device_alloc) \
-                                  -> Mapping[str, Any]:
+    async def generate_docker_args(
+        self,
+        docker: Docker,
+        device_alloc,
+    ) -> Mapping[str, Any]:
         memory = sum(device_alloc['mem'].values())
         return {
             'HostConfig': {
@@ -518,25 +545,29 @@ class MemoryPlugin(AbstractComputePlugin):
             }
         }
 
-    @classmethod
-    async def restore_from_container(cls, container: Container,
-                                     alloc_map: AbstractAllocMap) -> None:
+    async def restore_from_container(
+        self,
+        container: Container,
+        alloc_map: AbstractAllocMap,
+    ) -> None:
         assert isinstance(alloc_map, DiscretePropertyAllocMap)
         memory_limit = container.backend_obj['HostConfig']['Memory']
         alloc_map.apply_allocation({
             SlotName('mem'): {DeviceId('root'): memory_limit},
         })
 
-    @classmethod
-    async def get_attached_devices(cls, device_alloc: Mapping[SlotName, Mapping[DeviceId, Decimal]]) \
-                                   -> Sequence[DeviceModelInfo]:
+    async def get_attached_devices(
+        self,
+        device_alloc: Mapping[SlotName, Mapping[DeviceId, Decimal]],
+    ) -> Sequence[DeviceModelInfo]:
         device_ids = [*device_alloc[SlotName('mem')].keys()]
-        available_devices = await cls.list_devices()
+        available_devices = await self.list_devices()
         attached_devices: List[DeviceModelInfo] = []
         for device in available_devices:
             if device.device_id in device_ids:
                 attached_devices.append({
                     'device_id': device.device_id,
                     'model_name': '',
+                    'data': {},
                 })
         return attached_devices
