@@ -1,5 +1,6 @@
 import asyncio
 from decimal import Decimal
+import hashlib
 import io
 import ipaddress
 import json
@@ -8,9 +9,9 @@ from pathlib import Path
 import re
 from typing import (
     Any, Optional,
-    Callable, Iterable,
+    Iterable,
     Mapping, MutableMapping,
-    List, Sequence, Union,
+    List, Sequence, Tuple, Union,
     Type, overload,
     Set,
 )
@@ -28,7 +29,9 @@ from ai.backend.common.logging import BraceStyleAdapter
 from ai.backend.common.types import (
     PID, HostPID, ContainerPID, KernelId,
     ServicePort,
+    ServicePortProtocols,
 )
+from ai.backend.common.utils import current_loop
 
 log = BraceStyleAdapter(logging.getLogger('ai.backend.agent.utils'))
 
@@ -39,7 +42,12 @@ InOtherContainerPID: Final = ContainerPID(PID(-2))
 NotContainerPID: Final = ContainerPID(PID(-1))
 NotHostPID: Final = HostPID(PID(-1))
 
-_rx_service_ports = re.compile(r'^(?P<name>\w+):(?P<proto>\w+):(?P<ports>\[\d+(?:,\d+)*\]|\d+)(?:,|$)')
+_rx_service_ports = re.compile(
+    r'^(?P<name>[\w-]+):(?P<proto>\w+):(?P<ports>\[\d+(?:,\d+)*\]|\d+)(?:,|$)')
+
+
+def generate_agent_id(hint: str) -> str:
+    return hashlib.md5(hint.encode('utf-8')).hexdigest()[:12]
 
 
 def update_nested_dict(dest: MutableMapping, additions: Mapping) -> None:
@@ -57,13 +65,6 @@ def update_nested_dict(dest: MutableMapping, additions: Mapping) -> None:
                 dest[k] = v
 
 
-current_loop: Callable[[], asyncio.AbstractEventLoop]
-if hasattr(asyncio, 'get_running_loop'):
-    current_loop = asyncio.get_running_loop  # type: ignore
-else:
-    current_loop = asyncio.get_event_loop    # type: ignore
-
-
 def numeric_list(s: str) -> List[int]:
     return [int(p) for p in s.split()]
 
@@ -79,19 +80,24 @@ def parse_service_ports(s: str) -> Sequence[ServicePort]:
             if not name:
                 raise ValueError('Service port name must be not empty.')
             protocol = match.group('proto')
-            if protocol not in ('tcp', 'pty', 'http'):
+            if protocol == 'pty':
+                # unsupported, skip
+                continue
+            if protocol not in ('tcp', 'http', 'preopen'):
                 raise ValueError(f'Unsupported service port protocol: {protocol}')
             ports = tuple(map(int, match.group('ports').strip('[]').split(',')))
             for p in ports:
                 if p in used_ports:
                     raise ValueError(f'The port {p} is already used by another service port.')
+                if p >= 65535:
+                    raise ValueError(f'The service port number {p} must be smaller than 65535.')
                 if p in (2000, 2001, 2002, 2003, 2200, 7681):
                     raise ValueError('The service ports 2000 to 2003, 2200 and 7681 '
                                      'are reserved for internal use.')
                 used_ports.add(p)
             items.append({
                 'name': name,
-                'protocol': protocol,
+                'protocol': ServicePortProtocols(protocol),
                 'container_ports': ports,
                 'host_ports': (None,) * len(ports),
             })
@@ -140,7 +146,7 @@ def read_sysfs(path: Union[str, Path], type_: Type[Any], default: Any = None) ->
     try:
         raw_str = Path(path).read_text().strip()
         if type_ is bool:
-            return t.StrBool().check(raw_str)
+            return t.ToBool().check(raw_str)
         else:
             return type_(raw_str)
     except IOError:
@@ -196,10 +202,9 @@ async def get_subnet_ip(etcd: AsyncEtcd, network: str, fallback_addr: str = '0.0
 
 async def host_pid_to_container_pid(container_id: str, host_pid: HostPID) -> ContainerPID:
     kernel_ver = Path('/proc/version').read_text()
-    m = re.match(r'Linux version (\d+)\.(\d+)\..*', kernel_ver)
-    if m is not None:
-        matched_kernel_ver = tuple(m.groups())
-        if matched_kernel_ver < ('4', '1'):
+    if m := re.match(r'Linux version (\d+)\.(\d+)\..*', kernel_ver):  # noqa
+        kernel_ver_tuple: Tuple[str, str] = m.groups()  # type: ignore
+        if kernel_ver_tuple < ('4', '1'):
             # TODO: this should be deprecated when the minimun supported Linux kernel will be 4.1.
             #
             # In CentOs 7, NSPid is not accesible since it is supported from Linux kernel >=4.1.
@@ -284,16 +289,11 @@ async def host_pid_to_container_pid(container_id: str, host_pid: HostPID) -> Con
                     raise IndexError
                 container_pid = ContainerPID(container_table[process_idx][pid_idx])
                 log.debug('host pid {} is mapped to container pid {}', host_pid, container_pid)
-                return container_pid
+                return ContainerPID(PID(container_pid))
             except asyncio.CancelledError:
-                log.info('asyncio cancellederror')
                 raise
-            except (IndexError, KeyError, aiodocker.exceptions.DockerError) as e:
-                log.info(repr(e))
+            except (IndexError, KeyError, aiodocker.exceptions.DockerError):
                 return NotContainerPID
-            except Exception as e:
-                log.info('other exception')
-                log.info(repr(e))
             finally:
                 await docker.close()
 
