@@ -133,7 +133,7 @@ async def handle_fstab_detail(request: web.Request) -> web.Response:
     log.info('HANDLE_FSTAB_DETAIL')
     params = request.query
     fstab_path = params.get('fstab_path', '/etc/fstab')
-    async with aiofiles.open(fstab_path, mode='r') as fp:
+    async with aiofiles.open(fstab_path, mode='r') as fp:  # type: ignore
         content = await fp.read()
         return web.Response(text=content)
 
@@ -180,7 +180,7 @@ async def handle_mount(request: web.Request) -> web.Response:
     log.info('Mounted ' + params['name'] + ' on ' + mount_prefix)
     if params['edit_fstab']:
         fstab_path = params['fstab_path'] if params['fstab_path'] else '/etc/fstab'
-        async with aiofiles.open(fstab_path, mode='r+') as fp:
+        async with aiofiles.open(fstab_path, mode='r+') as fp:  # type: ignore
             fstab = Fstab(fp)
             await fstab.add(params['fs_location'], str(mountpoint),
                             params['fs_type'], params['options'])
@@ -213,7 +213,7 @@ async def handle_umount(request: web.Request) -> web.Response:
         pass
     if params['edit_fstab']:
         fstab_path = params['fstab_path'] if params['fstab_path'] else '/etc/fstab'
-        async with aiofiles.open(fstab_path, mode='r+') as fp:
+        async with aiofiles.open(fstab_path, mode='r+') as fp:  # type: ignore
             fstab = Fstab(fp)
             await fstab.remove_by_mountpoint(str(mountpoint))
     return web.Response(text=out)
@@ -354,7 +354,11 @@ def main(cli_ctx, config_path, debug):
         raise click.Abort()
 
     # Change the filename from the logging config's file section.
-    logger = Logger(cfg['logging'])
+    log_sockpath = Path(f'/tmp/backend.ai/ipc/watcher-logger-{os.getpid()}.sock')
+    log_sockpath.parent.mkdir(parents=True, exist_ok=True)
+    log_endpoint = f'ipc://{log_sockpath}'
+    cfg['logging']['endpoint'] = log_endpoint
+    logger = Logger(cfg['logging'], is_master=True, log_endpoint=log_endpoint)
     if 'file' in cfg['logging']['drivers']:
         fn = Path(cfg['logging']['file']['filename'])
         cfg['logging']['file']['filename'] = f"{fn.stem}-watcher{fn.suffix}"
