@@ -1,3 +1,4 @@
+from __future__ import annotations
 from abc import ABCMeta, abstractmethod
 import asyncio
 import concurrent.futures
@@ -12,7 +13,8 @@ import time
 from typing import (
     ClassVar,
     Optional, Union,
-    Sequence,
+    Awaitable,
+    Sequence, List,
     MutableMapping, Dict,
 )
 
@@ -22,10 +24,11 @@ from jupyter_client.kernelspec import KernelSpecManager
 import msgpack
 import zmq
 
+from .service import ServiceParser
 from .jupyter_client import aexecute_interactive
 from .logging import BraceStyleAdapter, setup_logger
-from .compat import asyncio_run_forever, current_loop
 from .utils import wait_local_port_open
+from .compat import current_loop
 from .intrinsic import (
     init_sshd_service, prepare_sshd_service,
     prepare_ttyd_service,
@@ -57,18 +60,23 @@ async def pipe_output(stream, outsock, target, log_fd):
         log.exception('unexpected error')
 
 
-async def terminate_and_kill(proc):
-    proc.terminate()
+async def terminate_and_wait(proc: asyncio.subprocess.Process) -> None:
     try:
-        await asyncio.wait_for(proc.wait(), timeout=2.0)
-    except asyncio.TimeoutError:
-        proc.kill()
-        await proc.wait()
+        proc.terminate()
+        try:
+            await asyncio.wait_for(proc.wait(), timeout=2.0)
+        except asyncio.TimeoutError:
+            proc.kill()
+            await proc.wait()
+    except ProcessLookupError:
+        pass
 
 
 class BaseRunner(metaclass=ABCMeta):
 
     log_prefix: ClassVar[str] = 'generic-kernel'
+    log_queue: janus.Queue[logging.LogRecord]
+    task_queue: asyncio.Queue[Awaitable[None]]
     default_runtime_path: ClassVar[Optional[str]] = None
     default_child_env: ClassVar[MutableMapping[str, str]] = {
         'LANG': 'C.UTF-8',
@@ -83,15 +91,22 @@ class BaseRunner(metaclass=ABCMeta):
 
     child_env: MutableMapping[str, str]
     subproc: Optional[asyncio.subprocess.Process]
-    runtime_path: Optional[Path]
+    service_parser: Optional[ServiceParser]
+    runtime_path: Path
 
     services_running: Dict[str, asyncio.subprocess.Process]
 
-    def __init__(self, loop=None):
+    _build_success: Optional[bool]
+
+    # Set by subclasses.
+    user_input_queue: Optional[asyncio.Queue[str]]
+
+    def __init__(self, runtime_path: Path) -> None:
         self.subproc = None
-        self.runtime_path = None
+        self.runtime_path = runtime_path
 
         self.child_env = {**os.environ, **self.default_child_env}
+        self._service_lock = asyncio.Lock()
         config_dir = Path('/home/config')
         try:
             evdata = (config_dir / 'environ.txt').read_text()
@@ -104,15 +119,7 @@ class BaseRunner(metaclass=ABCMeta):
         except Exception:
             log.exception('Reading /home/config/environ.txt failed!')
 
-        # initialized after loop creation
-        self.zctx = zmq.asyncio.Context()
         self.started_at: float = time.monotonic()
-        self.insock = None
-        self.outsock = None
-        self.init_done = None
-        self.task_queue = None
-        self.log_queue = None
-
         self.services_running = {}
 
         # If the subclass implements interatcive user inputs, it should set a
@@ -123,23 +130,69 @@ class BaseRunner(metaclass=ABCMeta):
         # build status tracker to skip the execute step
         self._build_success = None
 
-    async def _init_with_loop(self):
-        if self.init_done is not None:
-            self.init_done.clear()
+    async def _init(self, cmdargs) -> None:
+        self.cmdargs = cmdargs
+        loop = current_loop()
+        # Initialize event loop.
+        executor = concurrent.futures.ThreadPoolExecutor(max_workers=2)
+        loop.set_default_executor(executor)
+
+        self.zctx = zmq.asyncio.Context()
+        self.insock = self.zctx.socket(zmq.PULL)
+        self.insock.bind('tcp://*:2000')
+        self.outsock = self.zctx.socket(zmq.PUSH)
+        self.outsock.bind('tcp://*:2001')
+
+        self.log_queue = janus.Queue()
+        self.task_queue = asyncio.Queue()
+        self.init_done = asyncio.Event()
+
+        setup_logger(self.log_queue.sync_q, self.log_prefix, cmdargs.debug)
+        self._log_task = loop.create_task(self._handle_logs())
+        await asyncio.sleep(0)
+
+        service_def_folder = Path('/etc/backend.ai/service-defs')
+        if service_def_folder.is_dir():
+            self.service_parser = ServiceParser({
+                'runtime_path': str(self.runtime_path),
+            })
+            await self.service_parser.parse(service_def_folder)
+            log.debug('Loaded new-style service definitions.')
+        else:
+            self.service_parser = None
+
+        self._main_task = loop.create_task(self.main_loop(cmdargs))
+        self._run_task = loop.create_task(self.run_tasks())
+
+    async def _shutdown(self) -> None:
         try:
-            await self.init_with_loop()
-        except Exception:
-            log.exception('unexpected error')
-            return
-        await init_sshd_service(self.child_env)
-        if self.init_done is not None:
-            self.init_done.set()
+            self.insock.close()
+            log.debug('shutting down...')
+            self._run_task.cancel()
+            self._main_task.cancel()
+            await self._run_task
+            await self._main_task
+            log.debug('terminating service processes...')
+            running_procs = [*self.services_running.values()]
+            async with self._service_lock:
+                await asyncio.gather(
+                    *(terminate_and_wait(proc) for proc in running_procs),
+                    return_exceptions=True,
+                )
+                await asyncio.sleep(0.01)
+            log.debug('terminated.')
+        finally:
+            # allow remaining logs to be flushed.
+            await asyncio.sleep(0.1)
+            try:
+                if self.outsock:
+                    self.outsock.close()
+                await self._shutdown_jupyter_kernel()
+            finally:
+                self._log_task.cancel()
+                await self._log_task
 
-    @abstractmethod
-    async def init_with_loop(self):
-        """Initialize after the event loop is created."""
-
-    async def _init_jupyter_kernel(self):
+    async def _init_jupyter_kernel(self) -> None:
         """Detect ipython kernel spec for backend.ai and start it if found.
 
         Called after `init_with_loop`. `jupyter_kspec_name` should be defined to
@@ -179,7 +232,31 @@ class BaseRunner(metaclass=ABCMeta):
                       'no jupyter kernelspec found')
             self.kernel_mgr = None
 
-    async def _clean(self, clean_cmd):
+    async def _shutdown_jupyter_kernel(self):
+        if self.kernel_mgr and self.kernel_mgr.is_alive():
+            log.info('shutting down ' + self.jupyter_kspec_name + ' kernel...')
+            self.kernel_client.stop_channels()
+            self.kernel_mgr.shutdown_kernel()
+            assert not self.kernel_mgr.is_alive(), 'ipykernel failed to shutdown'
+
+    async def _init_with_loop(self) -> None:
+        if self.init_done is not None:
+            self.init_done.clear()
+        try:
+            await self.init_with_loop()
+            await init_sshd_service(self.child_env)
+        except Exception:
+            log.exception('Unexpected error!')
+            log.warning('We are skipping the error but the container may not work as expected.')
+            return
+        if self.init_done is not None:
+            self.init_done.set()
+
+    @abstractmethod
+    async def init_with_loop(self) -> None:
+        """Initialize after the event loop is created."""
+
+    async def _clean(self, clean_cmd: Optional[str]) -> None:
         ret = 0
         try:
             if clean_cmd is None or clean_cmd == '':
@@ -203,7 +280,19 @@ class BaseRunner(metaclass=ABCMeta):
         # it should not do anything by default.
         return 0
 
-    async def _build(self, build_cmd):
+    async def _bootstrap(self, script_path: Path) -> None:
+        log.info('Running the user bootstrap script...')
+        ret = 0
+        try:
+            ret = await self.run_subproc(['/bin/sh', str(script_path)])
+        except Exception:
+            log.exception('unexpected error while executing the user bootstrap script')
+            ret = -1
+        finally:
+            await asyncio.sleep(0.01)  # extra delay to flush logs
+            log.info('The user bootstrap script has exited with code {}', ret)
+
+    async def _build(self, build_cmd: Optional[str]) -> None:
         ret = 0
         try:
             if build_cmd is None or build_cmd == '':
@@ -231,7 +320,7 @@ class BaseRunner(metaclass=ABCMeta):
     async def build_heuristic(self) -> int:
         """Process build step."""
 
-    async def _execute(self, exec_cmd: str):
+    async def _execute(self, exec_cmd: str) -> None:
         ret = 0
         try:
             if exec_cmd is None or exec_cmd == '':
@@ -255,7 +344,7 @@ class BaseRunner(metaclass=ABCMeta):
     async def execute_heuristic(self) -> int:
         """Process execute step."""
 
-    async def _query(self, code_text):
+    async def _query(self, code_text: str) -> None:
         ret = 0
         try:
             ret = await self.query(code_text)
@@ -334,12 +423,12 @@ class BaseRunner(metaclass=ABCMeta):
                 await self.outsock.send_multipart(
                     [b'waiting-input',
                      json.dumps({'is_password': password}).encode('utf-8')])
-                user_input = await self._user_input_queue.async_q.get()
+                user_input = await self.user_input_queue.async_q.get()
                 self.kernel_client.input(user_input)
 
         # Run jupyter kernel's blocking execution method in an executor pool.
-        allow_stdin = False if self._user_input_queue is None else True
-        stdin_hook = None if self._user_input_queue is None else stdin_hook
+        allow_stdin = False if self.user_input_queue is None else True
+        stdin_hook = None if self.user_input_queue is None else stdin_hook  # type: ignore
         try:
             await aexecute_interactive(self.kernel_client, code_text, timeout=None,
                                        output_hook=output_hook,
@@ -350,15 +439,16 @@ class BaseRunner(metaclass=ABCMeta):
             return 127
         return 0
 
-    async def _complete(self, completion_data):
+    async def _complete(self, completion_data) -> Sequence[str]:
+        result: Sequence[str] = []
         try:
-            return await self.complete(completion_data)
+            result = await self.complete(completion_data)
         except Exception:
             log.exception('unexpected error')
         finally:
-            pass  # no need to "finished" signal
+            return result
 
-    async def complete(self, completion_data):
+    async def complete(self, completion_data) -> Sequence[str]:
         """Return the list of strings to be shown in the auto-complete list.
 
         The default interface is jupyter kernel. To use different interface,
@@ -414,7 +504,12 @@ class BaseRunner(metaclass=ABCMeta):
         return None, {}
 
     async def _start_service(self, service_info):
+        await self._service_lock.acquire()
         try:
+            if service_info['protocol'] == 'preopen':
+                # skip subprocess spawning as we assume the user runs it manually.
+                result = {'status': 'started'}
+                return
             if service_info['name'] in self.services_running:
                 result = {'status': 'running'}
                 return
@@ -423,16 +518,31 @@ class BaseRunner(metaclass=ABCMeta):
                           'error': 'not implemented yet'}
                 return
             cwd = Path.cwd()
-            cmdargs = None, {}
+            cmdargs, env = None, {}
             if service_info['name'] == 'ttyd':
                 cmdargs, env = await prepare_ttyd_service(service_info)
             elif service_info['name'] == 'sshd':
                 cmdargs, env = await prepare_sshd_service(service_info)
             elif service_info['name'] == 'vscode':
                 cmdargs, env = await prepare_vscode_service(service_info)
-            else:
-                cmdargs, env = await self.start_service(service_info)
+            elif self.service_parser is not None:
+                self.service_parser.variables['ports'] = service_info['ports']
+                cmdargs, env = await self.service_parser.start_service(
+                    service_info['name'],
+                    self.child_env.keys(),
+                    service_info['options'],
+                )
             if cmdargs is None:
+                # fall-back to legacy service routine
+                start_info = await self.start_service(service_info)
+                if start_info is None:
+                    cmdargs, env = None, {}
+                elif len(start_info) == 3:
+                    cmdargs, env, cwd = start_info
+                elif len(start_info) == 2:
+                    cmdargs, env = start_info
+            if cmdargs is None:
+                # still not found?
                 log.warning('The service {0} is not supported.',
                             service_info['name'])
                 result = {'status': 'failed',
@@ -445,19 +555,24 @@ class BaseRunner(metaclass=ABCMeta):
             if 'LD_LIBRARY_PATH' in service_env:
                 service_env['LD_LIBRARY_PATH'] = \
                     service_env['LD_LIBRARY_PATH'].replace('/opt/backend.ai/lib:', '')
-            loop = current_loop()
             try:
                 proc = await asyncio.create_subprocess_exec(
-                    *cmdargs,
+                    *map(str, cmdargs),
                     env=service_env,
                     cwd=cwd,
                 )
                 self.services_running[service_info['name']] = proc
-                loop.create_task(self._wait_service_proc(service_info['name'], proc))
+                asyncio.create_task(self._wait_service_proc(service_info['name'], proc))
+                self._service_lock.release()
                 await wait_local_port_open(service_info['port'])
                 log.info("Service {} has started (pid: {}, port: {})",
                          service_info['name'], proc.pid, service_info['port'])
                 result = {'status': 'started'}
+            except asyncio.CancelledError:
+                # This may happen if the service process gets started but it fails to
+                # open the port and then terminates (with an error).
+                result = {'status': 'failed',
+                          'error': f"the process did not start properly: {cmdargs[0]}"}
             except PermissionError:
                 result = {'status': 'failed',
                           'error': f"the target file is not executable: {cmdargs[0]}"}
@@ -465,13 +580,15 @@ class BaseRunner(metaclass=ABCMeta):
                 result = {'status': 'failed',
                           'error': f"the executable file is not found: {cmdargs[0]}"}
         except Exception as e:
-            log.exception('unexpected error')
+            log.exception('start_service: unexpected error')
             result = {'status': 'failed', 'error': repr(e)}
         finally:
             await self.outsock.send_multipart([
                 b'service-result',
                 json.dumps(result).encode('utf8'),
             ])
+            if self._service_lock.locked():
+                self._service_lock.release()
 
     async def _wait_service_proc(
         self,
@@ -482,7 +599,7 @@ class BaseRunner(metaclass=ABCMeta):
         log.info(f"Service {service_name} (pid: {proc.pid}) has terminated with exit code: {exitcode}")
         self.services_running.pop(service_name, None)
 
-    async def run_subproc(self, cmd: Sequence[Union[str, Path, None]], batch: bool = False):
+    async def run_subproc(self, cmd: Union[str, List[str]], batch: bool = False):
         """A thin wrapper for an external command."""
         loop = current_loop()
         if Path('/home/work/.logs').is_dir():
@@ -580,12 +697,30 @@ class BaseRunner(metaclass=ABCMeta):
             self.log_queue.close()
             await self.log_queue.wait_closed()
 
+    async def _get_apps(self, service_name):
+        result = {'status': 'done', 'data': []}
+        if self.service_parser is not None:
+            if service_name:
+                apps = await self.service_parser.get_apps(selected_service=service_name)
+            else:
+                apps = await self.service_parser.get_apps()
+            result['data'] = apps
+        await self.outsock.send_multipart([
+            b'apps-result',
+            json.dumps(result).encode('utf8'),
+        ])
+
     async def main_loop(self, cmdargs):
         user_input_server = \
             await asyncio.start_server(self.handle_user_input,
                                        '127.0.0.1', 65000)
         await self._init_with_loop()
         await self._init_jupyter_kernel()
+
+        user_bootstrap_path = Path('/home/work/bootstrap.sh')
+        if user_bootstrap_path.is_file():
+            await self._bootstrap(user_bootstrap_path)
+
         log.debug('start serving...')
         while True:
             try:
@@ -615,91 +750,18 @@ class BaseRunner(metaclass=ABCMeta):
                     await self._send_status()
                 elif op_type == 'start-service':  # activate a service port
                     data = json.loads(text)
-                    async with self._service_lock:
-                        await self._start_service(data)
+                    asyncio.create_task(self._start_service(data))
+                elif op_type == 'get-apps':
+                    await self._get_apps(text)
             except asyncio.CancelledError:
                 break
             except NotImplementedError:
                 log.error('Unsupported operation for this kernel: {0}', op_type)
                 await asyncio.sleep(0)
             except Exception:
-                log.exception('unexpected error')
+                log.exception('main_loop: unexpected error')
                 # we need to continue anyway unless we are shutting down
                 continue
         user_input_server.close()
         await user_input_server.wait_closed()
         await self.shutdown()
-
-    async def _init(self, cmdargs):
-        self.cmdargs = cmdargs
-        # Initialize event loop.
-        loop = current_loop()
-        executor = concurrent.futures.ThreadPoolExecutor(max_workers=2)
-        loop.set_default_executor(executor)
-        self.loop = loop
-
-        self.insock = self.zctx.socket(zmq.PULL)
-        self.insock.bind('tcp://*:2000')
-        self.outsock = self.zctx.socket(zmq.PUSH)
-        self.outsock.bind('tcp://*:2001')
-
-        self.log_queue = janus.Queue()
-        self.task_queue = asyncio.Queue()
-        self.init_done = asyncio.Event()
-
-        setup_logger(self.log_queue.sync_q, self.log_prefix, cmdargs.debug)
-        self._log_task = loop.create_task(self._handle_logs())
-        self._main_task = loop.create_task(self.main_loop(cmdargs))
-        self._run_task = loop.create_task(self.run_tasks())
-        self._service_lock = asyncio.Lock()
-
-    async def _shutdown(self):
-        try:
-            self.insock.close()
-            log.debug('shutting down...')
-            self._run_task.cancel()
-            self._main_task.cancel()
-            await self._run_task
-            await self._main_task
-            log.debug('terminating service processes...')
-            running_procs = [*self.services_running.values()]
-            async with self._service_lock:
-                await asyncio.gather(
-                    *(terminate_and_kill(proc) for proc in running_procs),
-                    return_exceptions=True,
-                )
-                await asyncio.sleep(0.01)
-            log.debug('terminated.')
-        finally:
-            # allow remaining logs to be flushed.
-            await asyncio.sleep(0.1)
-            try:
-                if self.outsock:
-                    self.outsock.close()
-                await self._shutdown_jupyter_kernel()
-            finally:
-                self._log_task.cancel()
-                await self._log_task
-
-    async def _shutdown_jupyter_kernel(self):
-        if self.kernel_mgr and self.kernel_mgr.is_alive():
-            log.info('shutting down ' + self.jupyter_kspec_name + ' kernel...')
-            self.kernel_client.stop_channels()
-            self.kernel_mgr.shutdown_kernel()
-            assert not self.kernel_mgr.is_alive(), 'ipykernel failed to shutdown'
-
-    def run(self, cmdargs):
-        if cmdargs.runtime_path is None:
-            self.runtime_path = self.default_runtime_path
-        else:
-            self.runtime_path = cmdargs.runtime_path
-
-        # Replace stdin with a "null" file
-        # (trying to read stdin will raise EOFError immediately afterwards.)
-        sys.stdin = open(os.devnull, 'rb')
-
-        # Terminal does not work with uvloop! :(
-        # FIXME: asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
-
-        asyncio_run_forever(self._init(cmdargs), self._shutdown(),
-                            stop_signals={signal.SIGINT, signal.SIGTERM})
